"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BuerliGeometryInteraction = void 0;
const core_1 = require("@buerli.io/core");
const fiber_1 = require("@react-three/fiber");
const react_1 = __importDefault(require("react"));
const CameraHelper_1 = require("../helpers/CameraHelper");
const findBuerliGeometry_1 = require("../helpers/findBuerliGeometry");
const hooks_1 = require("../hooks");
const Entity_1 = require("./geometry/overlay/Entity");
const Loop_1 = require("./geometry/overlay/Loop");
const Mesh_1 = require("./geometry/overlay/Mesh");
const Point_1 = require("./geometry/overlay/Point");
const Spline_1 = require("./geometry/overlay/Spline");
const GlobalTransform_1 = require("./GlobalTransform");
// TODO: - Everything related to "Global" selection is just a prototyp for now. dm,08/04/2022
//       - Make GlobalTransform optional, by default true
const LoopTypes = [core_1.GraphicType.LOOP];
const LineTypes = core_1.EdgeTypes.concat(core_1.CurveTypes);
const isLine = (type) => LineTypes.includes(type);
const isPoint = (type) => core_1.PointTypes.includes(type);
const isMesh = (type) => core_1.MeshTypes.includes(type);
const isLoop = (type) => LoopTypes.includes(type);
const isEntity = (type) => core_1.EntityTypes.includes(type);
const getKey = (elem) => `${elem.graphicId}|${elem.productId}|${elem.type}`;
exports.BuerliGeometryInteraction = react_1.default.memo(props => {
    const { drawingId, children, lineWidth = 4, pointSize = 6 } = props;
    const highlightColor = (0, hooks_1.useBuerli)(state => state.options.theme.highlightedGeom);
    const hoverColor = (0, hooks_1.useBuerli)(state => state.options.theme.hoveredGeom);
    const activeSelId = (0, hooks_1.useDrawing)(drawingId, d => d.selection.active);
    const activeItems = (0, hooks_1.useDrawing)(drawingId, d => { var _a; return (_a = d.selection.refs[activeSelId]) === null || _a === void 0 ? void 0 : _a.items; });
    const selApi = (0, hooks_1.useDrawing)(drawingId, d => d.api.selection);
    const lnTh = (0, fiber_1.useThree)(state => { var _a; return (_a = state.raycaster.params.Line) === null || _a === void 0 ? void 0 : _a.threshold; });
    const ptsTh = (0, fiber_1.useThree)(state => { var _a; return (_a = state.raycaster.params.Points) === null || _a === void 0 ? void 0 : _a.threshold; });
    const { camera, size } = (0, fiber_1.useThree)();
    const { lineThreshold, pointThreshold } = react_1.default.useMemo(() => {
        return {
            lineThreshold: lnTh || CameraHelper_1.CameraHelper.calculateScaleFactor(camera.position, lineWidth, camera, size),
            pointThreshold: ptsTh || CameraHelper_1.CameraHelper.calculateScaleFactor(camera.position, pointSize, camera, size),
        };
    }, [camera, lineWidth, pointSize, size, lnTh, ptsTh]);
    const [hovered, setHovered] = react_1.default.useState(null);
    /**
     * Stores a local copy of the selected elements. If a selector is active, it
     * contains the same items as the selector. If no selector is active, it contains
     * all items selected without a selector activated ("Global").
     */
    const [selected, setSelected] = react_1.default.useState([]);
    /**
     * Sync selected items on activeItems changed.
     */
    react_1.default.useEffect(() => {
        setSelected((activeItems === null || activeItems === void 0 ? void 0 : activeItems.filter(i => i.scope === core_1.BuerliScope).map(i => i.data)) || []);
    }, [setSelected, setHovered, activeItems]);
    /**
     * Remove the hovered item, if the active selector changes. This ensures the filter of the
     * new active selector will be triggered, even if already an item was hovered.
     */
    react_1.default.useEffect(() => {
        setHovered(null);
    }, [setHovered, activeSelId]);
    /**
     * Checks if the element is selectable and returns the element again. It also does some
     * special handling for loops and entities.
     *
     * Since entities and loops do not exist as an object, the item will be changed to be a
     * loop or entity if the active selector would like to be able to select loops or entiteis.
     */
    const getSelectable = react_1.default.useCallback((e) => {
        const elem = (0, findBuerliGeometry_1.findBuerliGeometry)(e, lineThreshold, pointThreshold)[0];
        if (!elem)
            return null;
        // if (!activeSelId) return elem // Everything is selectable if no selector is active (global selection)
        const selection = (0, core_1.getDrawing)(drawingId).selection.refs[activeSelId];
        if (selection === null || selection === void 0 ? void 0 : selection.isSelectable(core_1.BuerliScope, elem.container.type)) {
            // Use the container id of the element as graphicId for entity selection.
            // All elements of one entity must have the same graphicId in order to have
            // the entity selection working proper.
            return Object.assign(Object.assign({}, elem), { type: elem.container.type, graphicId: elem.container.id });
        }
        else if ((selection === null || selection === void 0 ? void 0 : selection.isSelectable(core_1.BuerliScope, core_1.GraphicType.LOOP)) && core_1.MeshTypes.indexOf(elem.type) >= 0) {
            // Special handling for LOOP's
            return Object.assign(Object.assign({}, elem), { type: core_1.GraphicType.LOOP });
        }
        else if (selection === null || selection === void 0 ? void 0 : selection.isSelectable(core_1.BuerliScope, elem.type)) {
            return elem;
        }
        return null;
    }, [activeSelId, drawingId, lineThreshold, pointThreshold]);
    /**
     * Selects the ProductElement or deselects it if it was already selected.
     */
    const selectOrDeselect = react_1.default.useCallback((elements) => {
        if (activeSelId) {
            // Specific selector is active
            const items = elements.map(s => (0, core_1.createGraphicItem)(s.productId, s));
            const haveUnselected = items.find(item => !(selApi === null || selApi === void 0 ? void 0 : selApi.isItemSelected(item))) !== undefined;
            if (haveUnselected) {
                selApi === null || selApi === void 0 ? void 0 : selApi.select(items, activeSelId);
            }
            else {
                selApi === null || selApi === void 0 ? void 0 : selApi.unselect(items, activeSelId);
            }
        }
        else {
            // TODO: handle "Global" selection
            // "Global" selection is active
            // let sel_ = [...selected]
            // for (const elem of elements) {
            //   // TODO: Is not yet a clean solution!
            //   const filtered = sel_.filter(s => getKey(s) != getKey(elem))
            //   filtered.length < sel_.length ? void (sel_ = filtered) : void (sel_ = [...sel_, elem])
            // }
            // setSelected(sel_)
        }
    }, [activeSelId, selApi /*selected*/]);
    /**
     * Handle click events on hovered objects.
     */
    const onClick = react_1.default.useCallback((e) => {
        if (e.delta > 4)
            return;
        const elem = getSelectable(e);
        if (elem) {
            e.stopPropagation();
            if (isLoop(elem.type)) {
                // Since we do not have loops in the model at all, we need to select all edges of the loop instead of the not exiting loop.
                const edges = [];
                const anyElem = elem;
                if (anyElem && anyElem.loops) {
                    const loops = anyElem.loops;
                    for (const loop of loops) {
                        edges.push(...loop);
                    }
                }
                const container = (0, core_1.getDrawing)(drawingId).geometry.cache[elem.container.id];
                const elems = container ? edges.map(n => (Object.assign(Object.assign({}, container.map[n]), { productId: elem.productId }))) : [];
                selectOrDeselect(elems);
            }
            else {
                selectOrDeselect([elem]);
            }
        }
    }, [drawingId, getSelectable, selectOrDeselect]);
    /**
     * Handle over events. Searches buerli geometry in raycast intersections.
     */
    const onOver = react_1.default.useCallback((e) => setHovered(getSelectable(e)), [getSelectable]);
    /**
     * Handle out events. Searches buerli geometry in raycast intersections.
     */
    const onOut = react_1.default.useCallback((e) => setHovered(getSelectable(e)), [getSelectable]);
    /**
     * Register handlers only if a selection is active.
     */
    const handlers = activeSelId ? { onPointerMove: onOver, onPointerOut: onOut, onClick: onClick } : {};
    return (react_1.default.createElement("group", Object.assign({}, handlers),
        children,
        selected.map(sel => (react_1.default.createElement(GlobalTransform_1.GlobalTransform, { key: getKey(sel), drawingId: drawingId, objectId: sel.productId },
            isMesh(sel.type) && react_1.default.createElement(Mesh_1.Mesh, { elem: sel, renderOrder: 500, color: highlightColor }),
            isLine(sel.type) && react_1.default.createElement(Spline_1.Spline, { elem: sel, renderOrder: 510, color: highlightColor }),
            isPoint(sel.type) && react_1.default.createElement(Point_1.Point, { elem: sel, renderOrder: 520, color: highlightColor }),
            isEntity(sel.type) && react_1.default.createElement(Entity_1.Entity, { drawingId: drawingId, elem: sel, renderOrder: 530, color: highlightColor })))),
        hovered && (react_1.default.createElement(GlobalTransform_1.GlobalTransform, { drawingId: drawingId, objectId: hovered.productId },
            isMesh(hovered.type) && react_1.default.createElement(Mesh_1.Mesh, { elem: hovered, renderOrder: 501, color: hoverColor }),
            isLine(hovered.type) && react_1.default.createElement(Spline_1.Spline, { elem: hovered, renderOrder: 511, color: hoverColor, lineWidth: 6.1 }),
            isPoint(hovered.type) && react_1.default.createElement(Point_1.Point, { elem: hovered, renderOrder: 521, color: hoverColor, pointSize: 6.1 }),
            isLoop(hovered.type) && react_1.default.createElement(Loop_1.Loop, { drawingId: drawingId, elem: hovered, renderOrder: 531, color: hoverColor }),
            isEntity(hovered.type) && (react_1.default.createElement(Entity_1.Entity, { drawingId: drawingId, elem: hovered, renderOrder: 532, color: hoverColor }))))));
});
