"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.useHeadless = exports.headless = void 0;
const react_1 = require("react");
const classcad_1 = require("@buerli.io/classcad");
const suspend_react_1 = require("suspend-react");
/**
 * Creates a global ClassCAD connection and returns a set of functions to interact with it.
 * Use this if you want to create a single headless instance that is shared across the entire app.
 * @param {solid|history} impl The desired headless instance, for example: solid or history
 * @param {string} url The url of the ClassCAD server
 * @param {HeadlessConfig} config Optional object of options
 *  - socket: Buerli socket adapter, default: SocketIOClient
    - lifespan: Suspense entry invalidation in ms, default: 0 (keep-alive forever)
    - equal: Suspense cache equality function, default: (a, b) => a === b (reference equality)
    - store: An object that can persist values for the lenght of the session, default: {}
 * @returns {HeadlessReturn} A set of functions to interact with the headless instance
 * @example
 * const { run } = headless(solid, 'http://localhost:9091')
 * function App() {
 *   useEffect(() => {
 *     run(async (api) => {
 *       const id = await api.createBox(100, 100, 100)
 *     })
 *   }, [])
 *   return <BuerliGeometry />
 * }
 */
const headless = (impl, url, _a = {}) => {
    var { store, socket, lifespan, equal } = _a, config = __rest(_a, ["store", "socket", "lifespan", "equal"]);
    if (!store)
        store = {};
    (0, classcad_1.init)(id => new (socket || classcad_1.SocketIOClient)(url, id), config);
    const instance = new impl();
    const instanceApi = new Promise(resolve => instance.init(api => resolve(api)));
    return {
        store,
        instance,
        api: instanceApi,
        useDrawingId: () => (0, suspend_react_1.suspend)(() => instanceApi.then(() => instance.drawingId), [instance]),
        run: async (callback) => callback(await instanceApi, store),
        cache: (callback, dependencies) => (0, suspend_react_1.suspend)(async (...keys) => callback(await instanceApi, store, ...keys), dependencies, { lifespan, equal }),
        peek: (dependencies) => (0, suspend_react_1.peek)(dependencies),
        preload: (callback, dependencies) => (0, suspend_react_1.preload)(async (...keys) => callback(await instanceApi, ...keys), dependencies),
    };
};
exports.headless = headless;
/**
 * Creates a component-bound ClassCAD connection and returns a set of functions to interact with it.
 * Multiple components using this hook with the same impl and url will access the same headless instance.
 * Once the last component using it unmounts, the headless instance will be destroyed.
 * If you want unique per-component headless instances, add a postfix to the url: 'ws://localhost:9091?session=mainSession'
 * @param {solid|history} impl The desired headless instance, for example: solid or history
 * @param {string} url The url of the ClassCAD server
 * @param {HeadlessConfig} config Optional object of options
 *  - socket: Buerli socket adapter, default: SocketIOClient
    - lifespan: Suspense entry invalidation in ms, default: 0 (keep-alive forever)
    - equal: Suspense cache equality function, default: (a, b) => a === b (reference equality)
    - store: An object that can persist values for the lenght of the session, default: {}
 * @returns {HeadlessReturn} A set of functions to interact with the headless instance
 * @example
 * function App() {
 *   const { run } = useHeadless(solid, 'ws://localhost:9091?session=mainSession')
 *   useEffect(() => {
 *     run(async (api) => {
 *       const id = await api.createBox(100, 100, 100)
 *     })
 *   }, [])
 *   return <BuerliGeometry />
 * }
 */
const useHeadless = (impl, url, config = {}) => {
    // Create a cached headless instance, the implementation and url act as the cache keys
    const buerli = (0, suspend_react_1.suspend)(async () => {
        const temp = (0, exports.headless)(impl, url, config);
        temp._increment = 0;
        return temp;
    }, [impl, url]);
    // Destroy the headless instance when the last component using it unmounts
    (0, react_1.useEffect)(() => {
        // Increment the counter when a component mounts
        buerli._increment++;
        return () => {
            // Decrement the counter when a component unmounts
            buerli._increment--;
            // Destroy the headless instance when the counter reaches 0
            if (buerli._increment <= 0) {
                // Destroy the headless instance
                buerli.instance.destroy();
                // Invalidate the cache entry
                (0, suspend_react_1.clear)([impl, url]);
            }
        };
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [buerli, impl, url]);
    return buerli;
};
exports.useHeadless = useHeadless;
