import { BuerliScope, createGraphicItem, CurveTypes, EdgeTypes, EntityTypes, getDrawing, GraphicType, MeshTypes, PointTypes, } from '@buerli.io/core';
import { useThree } from '@react-three/fiber';
import React from 'react';
import { CameraHelper } from '../helpers/CameraHelper';
import { findBuerliGeometry } from '../helpers/findBuerliGeometry';
import { useBuerli, useDrawing } from '../hooks';
import { Entity } from './geometry/overlay/Entity';
import { Loop } from './geometry/overlay/Loop';
import { Mesh } from './geometry/overlay/Mesh';
import { Point } from './geometry/overlay/Point';
import { Spline } from './geometry/overlay/Spline';
import { GlobalTransform } from './GlobalTransform';
// TODO: - Everything related to "Global" selection is just a prototyp for now. dm,08/04/2022
//       - Make GlobalTransform optional, by default true
const LoopTypes = [GraphicType.LOOP];
const LineTypes = EdgeTypes.concat(CurveTypes);
const isLine = (type) => LineTypes.includes(type);
const isPoint = (type) => PointTypes.includes(type);
const isMesh = (type) => MeshTypes.includes(type);
const isLoop = (type) => LoopTypes.includes(type);
const isEntity = (type) => EntityTypes.includes(type);
const getKey = (elem) => `${elem.graphicId}|${elem.productId}|${elem.type}`;
export const BuerliGeometryInteraction = React.memo(props => {
    const { drawingId, children, lineWidth = 4, pointSize = 6 } = props;
    const highlightColor = useBuerli(state => state.options.theme.highlightedGeom);
    const hoverColor = useBuerli(state => state.options.theme.hoveredGeom);
    const activeSelId = useDrawing(drawingId, d => d.selection.active);
    const activeItems = useDrawing(drawingId, d => { var _a; return (_a = d.selection.refs[activeSelId]) === null || _a === void 0 ? void 0 : _a.items; });
    const selApi = useDrawing(drawingId, d => d.api.selection);
    const lnTh = useThree(state => { var _a; return (_a = state.raycaster.params.Line) === null || _a === void 0 ? void 0 : _a.threshold; });
    const ptsTh = useThree(state => { var _a; return (_a = state.raycaster.params.Points) === null || _a === void 0 ? void 0 : _a.threshold; });
    const { camera, size } = useThree();
    const { lineThreshold, pointThreshold } = React.useMemo(() => {
        return {
            lineThreshold: lnTh || CameraHelper.calculateScaleFactor(camera.position, lineWidth, camera, size),
            pointThreshold: ptsTh || CameraHelper.calculateScaleFactor(camera.position, pointSize, camera, size),
        };
    }, [camera, lineWidth, pointSize, size, lnTh, ptsTh]);
    const [hovered, setHovered] = React.useState(null);
    /**
     * Stores a local copy of the selected elements. If a selector is active, it
     * contains the same items as the selector. If no selector is active, it contains
     * all items selected without a selector activated ("Global").
     */
    const [selected, setSelected] = React.useState([]);
    /**
     * Sync selected items on activeItems changed.
     */
    React.useEffect(() => {
        setSelected((activeItems === null || activeItems === void 0 ? void 0 : activeItems.filter(i => i.scope === BuerliScope).map(i => i.data)) || []);
    }, [setSelected, setHovered, activeItems]);
    /**
     * Remove the hovered item, if the active selector changes. This ensures the filter of the
     * new active selector will be triggered, even if already an item was hovered.
     */
    React.useEffect(() => {
        setHovered(null);
    }, [setHovered, activeSelId]);
    /**
     * Checks if the element is selectable and returns the element again. It also does some
     * special handling for loops and entities.
     *
     * Since entities and loops do not exist as an object, the item will be changed to be a
     * loop or entity if the active selector would like to be able to select loops or entiteis.
     */
    const getSelectable = React.useCallback((e) => {
        const elem = findBuerliGeometry(e, lineThreshold, pointThreshold)[0];
        if (!elem)
            return null;
        // if (!activeSelId) return elem // Everything is selectable if no selector is active (global selection)
        const selection = getDrawing(drawingId).selection.refs[activeSelId];
        if (selection === null || selection === void 0 ? void 0 : selection.isSelectable(BuerliScope, elem.container.type)) {
            // Use the container id of the element as graphicId for entity selection.
            // All elements of one entity must have the same graphicId in order to have
            // the entity selection working proper.
            return Object.assign(Object.assign({}, elem), { type: elem.container.type, graphicId: elem.container.id });
        }
        else if ((selection === null || selection === void 0 ? void 0 : selection.isSelectable(BuerliScope, GraphicType.LOOP)) && MeshTypes.indexOf(elem.type) >= 0) {
            // Special handling for LOOP's
            return Object.assign(Object.assign({}, elem), { type: GraphicType.LOOP });
        }
        else if (selection === null || selection === void 0 ? void 0 : selection.isSelectable(BuerliScope, elem.type)) {
            return elem;
        }
        return null;
    }, [activeSelId, drawingId, lineThreshold, pointThreshold]);
    /**
     * Selects the ProductElement or deselects it if it was already selected.
     */
    const selectOrDeselect = React.useCallback((elements) => {
        if (activeSelId) {
            // Specific selector is active
            const items = elements.map(s => createGraphicItem(s.productId, s));
            const haveUnselected = items.find(item => !(selApi === null || selApi === void 0 ? void 0 : selApi.isItemSelected(item))) !== undefined;
            if (haveUnselected) {
                selApi === null || selApi === void 0 ? void 0 : selApi.select(items, activeSelId);
            }
            else {
                selApi === null || selApi === void 0 ? void 0 : selApi.unselect(items, activeSelId);
            }
        }
        else {
            // TODO: handle "Global" selection
            // "Global" selection is active
            // let sel_ = [...selected]
            // for (const elem of elements) {
            //   // TODO: Is not yet a clean solution!
            //   const filtered = sel_.filter(s => getKey(s) != getKey(elem))
            //   filtered.length < sel_.length ? void (sel_ = filtered) : void (sel_ = [...sel_, elem])
            // }
            // setSelected(sel_)
        }
    }, [activeSelId, selApi /*selected*/]);
    /**
     * Handle click events on hovered objects.
     */
    const onClick = React.useCallback((e) => {
        if (e.delta > 4)
            return;
        const elem = getSelectable(e);
        if (elem) {
            e.stopPropagation();
            if (isLoop(elem.type)) {
                // Since we do not have loops in the model at all, we need to select all edges of the loop instead of the not exiting loop.
                const edges = [];
                const anyElem = elem;
                if (anyElem && anyElem.loops) {
                    const loops = anyElem.loops;
                    for (const loop of loops) {
                        edges.push(...loop);
                    }
                }
                const container = getDrawing(drawingId).geometry.cache[elem.container.id];
                const elems = container ? edges.map(n => (Object.assign(Object.assign({}, container.map[n]), { productId: elem.productId }))) : [];
                selectOrDeselect(elems);
            }
            else {
                selectOrDeselect([elem]);
            }
        }
    }, [drawingId, getSelectable, selectOrDeselect]);
    /**
     * Handle over events. Searches buerli geometry in raycast intersections.
     */
    const onOver = React.useCallback((e) => setHovered(getSelectable(e)), [getSelectable]);
    /**
     * Handle out events. Searches buerli geometry in raycast intersections.
     */
    const onOut = React.useCallback((e) => setHovered(getSelectable(e)), [getSelectable]);
    /**
     * Register handlers only if a selection is active.
     */
    const handlers = activeSelId ? { onPointerMove: onOver, onPointerOut: onOut, onClick: onClick } : {};
    return (React.createElement("group", Object.assign({}, handlers),
        children,
        selected.map(sel => (React.createElement(GlobalTransform, { key: getKey(sel), drawingId: drawingId, objectId: sel.productId },
            isMesh(sel.type) && React.createElement(Mesh, { elem: sel, renderOrder: 500, color: highlightColor }),
            isLine(sel.type) && React.createElement(Spline, { elem: sel, renderOrder: 510, color: highlightColor }),
            isPoint(sel.type) && React.createElement(Point, { elem: sel, renderOrder: 520, color: highlightColor }),
            isEntity(sel.type) && React.createElement(Entity, { drawingId: drawingId, elem: sel, renderOrder: 530, color: highlightColor })))),
        hovered && (React.createElement(GlobalTransform, { drawingId: drawingId, objectId: hovered.productId },
            isMesh(hovered.type) && React.createElement(Mesh, { elem: hovered, renderOrder: 501, color: hoverColor }),
            isLine(hovered.type) && React.createElement(Spline, { elem: hovered, renderOrder: 511, color: hoverColor, lineWidth: 6.1 }),
            isPoint(hovered.type) && React.createElement(Point, { elem: hovered, renderOrder: 521, color: hoverColor, pointSize: 6.1 }),
            isLoop(hovered.type) && React.createElement(Loop, { drawingId: drawingId, elem: hovered, renderOrder: 531, color: hoverColor }),
            isEntity(hovered.type) && (React.createElement(Entity, { drawingId: drawingId, elem: hovered, renderOrder: 532, color: hoverColor }))))));
});
