import * as THREE from 'three';
const viewMatrix = new THREE.Matrix4();
export const CameraHelper = {
    /**
     * Projects 2-dimensional coordinate from a 3-dimensional point within the view's scene.
     * @param {THREE.Vector3} point3 - Input point
     * @example
     * // Grab x and y off THREE's projected Vector2
     * let {x, y} = view.getPoint2D(new THREE.Vector3(10, 20, 100));
     * @returns {THREE.Vector2} The projected point
     */
    getPoint2(point3, camera, size) {
        const widthHalf = size.width / 2;
        const heightHalf = size.height / 2;
        // For some reasons camera matrix and it's inverse are not in sync here.
        // A performance optimisation (three 96: https://github.com/mrdoob/three.js/pull/14628) says that
        // both matrices are always in sync. If the user sets it manually he will have to set the inverse too.
        // See this: https://github.com/mrdoob/three.js/pull/14685#issuecomment-411982184
        // Somewhere in our code we set the matrix but not its inverse. Therefor we sync it here explicitly
        camera.updateMatrixWorld(false);
        const vector = point3.project(camera);
        vector.x = vector.x * widthHalf + widthHalf;
        vector.y = -(vector.y * heightHalf) + heightHalf;
        return vector;
    },
    /**
     * Projects a 2D point into 3D space. Z-Value for the 2D-Point can be passed within bounds of 0 to 1.
     * Note: maximal depth is used by default (z = 1), so the point is on the "far" frustum (with huge coordinates)
     */
    getPoint3(point2, camera, size, zValue = 1) {
        const vector = new THREE.Vector3((point2.x / size.width) * 2 - 1, -(point2.y / size.height) * 2 + 1, zValue);
        vector.unproject(camera);
        return vector;
    },
    /**
     * Calculate the scale factor that will make the object at the point3 position be radiusPx pixels wide.
     * Example use: object.scale.set(calculateScaleFactor(object.getWorldPosition(), radiusPx));
     * @param point3   Position of the object
     * @param radiusPx Desired radius of the object in pixels
     * @param camera   The camera
     * @param size     The size of the canvas
     * @return         Scale factor for the object
     */
    calculateScaleFactor(point3, radiusPx, camera, size) {
        const point2 = CameraHelper.getPoint2(point3.clone(), camera, size);
        let scale = 0;
        for (let i = 0; i < 2; ++i) {
            const point2off = point2.clone().setComponent(i, point2.getComponent(i) + radiusPx);
            const point3off = CameraHelper.getPoint3(point2off, camera, size, point2off.z);
            scale = Math.max(scale, point3.distanceTo(point3off));
        }
        return scale;
    },
    calculateNearFar: (camera, box) => {
        camera.updateMatrixWorld(false);
        // https://stackoverflow.com/questions/20510404/calculating-camera-near-and-far-distance-based-on-bounding-box
        viewMatrix.copy(camera.matrixWorld).invert();
        let minz = +Infinity;
        let maxz = -Infinity;
        for (let i = 0; i < 8; i++) {
            const x = i & 1 ? box.min.x : box.max.x;
            const y = i & 2 ? box.min.y : box.max.y;
            let z = i & 4 ? box.min.z : box.max.z;
            const p = new THREE.Vector3(x, y, z);
            p.applyMatrix4(viewMatrix);
            z = -p.z;
            if (z > maxz)
                maxz = z;
            if (z < minz)
                minz = z;
        }
        return { near: minz, far: maxz };
    },
};
