import { AwvNodeClient } from '@buerli.io/classcad';
import { BuerliState, DrawingID } from '@buerli.io/core';
import { Solid, History } from '@buerli.io/headless';
declare type Tuple<T = any> = [T] | T[];
declare type Await<T> = T extends Promise<infer V> ? V : never;
declare type InstanceType<T> = T extends typeof Solid ? Solid : History;
declare type ApiType<T> = T extends new () => {
    init: (callback: (api: infer U) => void) => () => Promise<void>;
} ? U : never;
declare type HeadlessConfig<Store> = Partial<BuerliState['options']> & {
    /** Buerli socket adapter, default: SocketIOClient */
    socket?: new (...args: any) => AwvNodeClient;
    /** Suspense entry invalidation in ms, default: 0 (keep-alive forever) */
    lifespan?: number;
    /** Suspense cache equality function, default: (a, b) => a === b (reference equality) */
    equal?: (a: any, b: any) => boolean;
    /** An object that can persist values for the lenght of the session */
    store?: Store;
};
declare type HeadlessReturn<Impl, Store> = {
    /** A user-defined store where variables can be shared */
    store: Store;
    /** The headless instance, for example: new solid() */
    instance: InstanceType<Impl>;
    /** The headless instance API, for example: new solid().init(api => ...) */
    api: Promise<ApiType<Impl>>;
    /**
     * A hook that suspends the component until the drawing ID is available
     * @returns {DrawingID} The drawing ID
     */
    useDrawingId: () => DrawingID;
    /**
     * Executes ClassCAD commands
     * @param {function} callback A function that receives the headless API as the first parameter,
     * and the user-defined store as the second
     * @example
     * function Box() {
     *   useEffect(() => {
     *     run(async (api) => {
     *       const id = await api.createBox(100, 100, 100)
     *     })
     *   }, [])
     *   return <BuerliGeometry />
     * }
     */
    run: <Fn extends (api: ApiType<Impl>, store: Store) => Promise<unknown>>(callback: Fn) => Promise<Await<ReturnType<Fn>>>;
    /**
     * Executes ClassCAD commands and returns a memoized result (using React suspense)
     * @param {function} callback A function that receives the headless API as the first parameter,
     * the user-defined store as the second, and the dependencies as the rest
     * @param {Array} dependencies An array of dependencies that will trigger a re-execution of the callback
     * @returns {any} The result of the callback
     * @example
     * function Box({ size }) {
     *   const geo = cache(async (api) => {
     *     const id = await api.createBox(size, size, size)
     *     return await api.createBufferGeometry(id)
     *   }, [size])
     *   return (
     *     <mesh geometry={geo}>
     *       <meshStandardMaterial />
     *     </mesh>
     *   )
     * }
     */
    cache: <Keys extends Tuple<unknown>, Fn extends (api: ApiType<Impl>, store: Store, ...keys: Keys) => Promise<unknown>>(callback: Fn, dependencies: Keys) => Await<ReturnType<Fn>>;
    /**
     * Looks up a previously memoized result (that was executed by the cache function above)
     * @param {Array} dependencies An array of dependencies
     * @returns {any} The result of the callback, if present
     */
    peek: <Keys extends Tuple<unknown>>(dependencies: Keys) => unknown;
    /**
     * Executes ClassCAD commands and preloads the memoized result
     * This function can be executed in global space to preload data before the app starts.
     * By the time cache(fn, deps) is called with the same dependencies, the data will be available immediately.
     * @param {function} callback A function that receives the headless API as the first parameter,
     * the user-defined store as the second, and the dependencies as the rest
     * @param {Array} dependencies An array of dependencies
     */
    preload: <Keys extends Tuple<unknown>, Fn extends (api: ApiType<Impl>, ...keys: Keys) => Promise<unknown>>(callback: Fn, dependencies: Keys) => unknown;
};
/**
 * Creates a global ClassCAD connection and returns a set of functions to interact with it.
 * Use this if you want to create a single headless instance that is shared across the entire app.
 * @param {solid|history} impl The desired headless instance, for example: solid or history
 * @param {string} url The url of the ClassCAD server
 * @param {HeadlessConfig} config Optional object of options
 *  - socket: Buerli socket adapter, default: SocketIOClient
    - lifespan: Suspense entry invalidation in ms, default: 0 (keep-alive forever)
    - equal: Suspense cache equality function, default: (a, b) => a === b (reference equality)
    - store: An object that can persist values for the lenght of the session, default: {}
 * @returns {HeadlessReturn} A set of functions to interact with the headless instance
 * @example
 * const { run } = headless(solid, 'http://localhost:9091')
 * function App() {
 *   useEffect(() => {
 *     run(async (api) => {
 *       const id = await api.createBox(100, 100, 100)
 *     })
 *   }, [])
 *   return <BuerliGeometry />
 * }
 */
export declare const headless: <Impl extends typeof Solid | typeof History, Store extends {
    [key: string]: any;
}>(impl: Impl, url: string, { store, socket, lifespan, equal, ...config }?: HeadlessConfig<Store>) => HeadlessReturn<Impl, Store>;
/**
 * Creates a component-bound ClassCAD connection and returns a set of functions to interact with it.
 * Multiple components using this hook with the same impl and url will access the same headless instance.
 * Once the last component using it unmounts, the headless instance will be destroyed.
 * If you want unique per-component headless instances, add a postfix to the url: 'ws://localhost:9091?session=mainSession'
 * @param {solid|history} impl The desired headless instance, for example: solid or history
 * @param {string} url The url of the ClassCAD server
 * @param {HeadlessConfig} config Optional object of options
 *  - socket: Buerli socket adapter, default: SocketIOClient
    - lifespan: Suspense entry invalidation in ms, default: 0 (keep-alive forever)
    - equal: Suspense cache equality function, default: (a, b) => a === b (reference equality)
    - store: An object that can persist values for the lenght of the session, default: {}
 * @returns {HeadlessReturn} A set of functions to interact with the headless instance
 * @example
 * function App() {
 *   const { run } = useHeadless(solid, 'ws://localhost:9091?session=mainSession')
 *   useEffect(() => {
 *     run(async (api) => {
 *       const id = await api.createBox(100, 100, 100)
 *     })
 *   }, [])
 *   return <BuerliGeometry />
 * }
 */
export declare const useHeadless: <Impl extends typeof Solid | typeof History, Store extends {
    [key: string]: any;
}>(impl: Impl, url: string, config?: HeadlessConfig<Store>) => Omit<HeadlessReturn<Impl, Store>, "_increment">;
export {};
